@@ -5,6 +5,7 @@ import (
 	"context"
 	"fmt"
 	"io"
+	"maps"
 	"os"
 	"path"
 	"path/filepath"
@@ -17,7 +18,10 @@ import (
 	"github.com/open-policy-agent/opa/internal/file/archive"
 	"github.com/open-policy-agent/opa/v1/ast"
 	"github.com/open-policy-agent/opa/v1/bundle"
+	"github.com/open-policy-agent/opa/v1/ir"
 	"github.com/open-policy-agent/opa/v1/rego"
+	"github.com/open-policy-agent/opa/v1/repl"
+	"github.com/open-policy-agent/opa/v1/storage/inmem"
 	"github.com/open-policy-agent/opa/v1/topdown"
 	"github.com/open-policy-agent/opa/v1/util/test"
 )
@@ -335,9 +339,9 @@ FAIL: 1/1
 				testParams.varValues = tc.includeVars
 				_ = testParams.explain.Set(explainModeFull)
 
-				_, err := opaTest([]string{root}, testParams)
-				if err != nil {
-					t.Fatalf("Unexpected error: %s", err)
+				errorCode := opaTest([]string{root}, testParams)
+				if errorCode != 2 {
+					t.Fatalf("Unexpected error code: %d", errorCode)
 				}
 
 				actual := buf.String()
@@ -1484,9 +1488,9 @@ FAIL: 1/1
 				testParams.varValues = true
 				_ = testParams.explain.Set(explainModeFull)
 
-				_, err := opaTest([]string{root}, testParams)
-				if err != nil {
-					t.Fatalf("Unexpected error: %s", err)
+				exitCode := opaTest([]string{root}, testParams)
+				if exitCode != 2 {
+					t.Fatalf("Unexpected error code: %d", exitCode)
 				}
 
 				actual := r.ReplaceAllString(buf.String(), "FAIL (%TIME%)")
@@ -1520,7 +1524,7 @@ test_p if {
 		testParams.bundleMode = false
 		testParams.ignore = []string{"broken.rego"}
 
-		exitCode, _ = opaTest([]string{root}, testParams)
+		exitCode = opaTest([]string{root}, testParams)
 	})
 
 	if exitCode > 0 {
@@ -1547,7 +1551,7 @@ test_p if {
 		testParams.errOutput = io.Discard
 		testParams.bundleMode = true
 		testParams.ignore = []string{"broken.rego"}
-		exitCode, _ = opaTest([]string{root}, testParams)
+		exitCode = opaTest([]string{root}, testParams)
 	})
 
 	if exitCode > 0 {
@@ -1555,24 +1559,24 @@ test_p if {
 	}
 }
 
-func testSchemasAnnotation(rego string) (int, error) {
+func testSchemasAnnotation(rego string) (int, []byte) {
 
 	files := map[string]string{
 		"test.rego": rego,
 	}
 
 	var exitCode int
-	var err error
+	var buf bytes.Buffer
 	test.WithTempFS(files, func(path string) {
 		regoFilePath := filepath.Join(path, "test.rego")
 
 		testParams := newTestCommandParams()
 		testParams.count = 1
-		testParams.errOutput = io.Discard
+		testParams.errOutput = &buf
 
-		exitCode, err = opaTest([]string{regoFilePath}, testParams)
+		exitCode = opaTest([]string{regoFilePath}, testParams)
 	})
-	return exitCode, err
+	return exitCode, buf.Bytes()
 }
 
 // Assert that 'schemas' annotations with schema ref are ignored, but not inlined schemas
@@ -1612,17 +1616,19 @@ test_p if {
     p with input.foo as 42
 }`
 
-	exitCode, err := testSchemasAnnotation(policyWithInlinedSchema)
+	exitCode, errOutput := testSchemasAnnotation(policyWithInlinedSchema)
 	// We expect an error here, as inlined schemas are always used for type checking
 
 	if exitCode == 0 {
 		t.Fatalf("didn't get expected error when inlined schema is present")
-	} else if !strings.Contains(err.Error(), "rego_type_error: match error") {
-		t.Fatalf("didn't get expected %s error when inlined schema is present; got: %v", ast.TypeErr, err)
+	}
+
+	if !strings.Contains(string(errOutput), "rego_type_error: match error") {
+		t.Fatalf("didn't get expected %s error when inlined schema is present; got: %v", ast.TypeErr, string(errOutput))
 	}
 }
 
-func testSchemasAnnotationWithJSONFile(rego string, schema string) (int, error) {
+func testSchemasAnnotationWithJSONFile(rego string, schema string) (int, []byte) {
 
 	files := map[string]string{
 		"test.rego":        rego,
@@ -1630,18 +1636,19 @@ func testSchemasAnnotationWithJSONFile(rego string, schema string) (int, error)
 	}
 
 	var exitCode int
-	var err error
+	var buf bytes.Buffer
 	test.WithTempFS(files, func(path string) {
 		regoFilePath := filepath.Join(path, "test.rego")
 
 		testParams := newTestCommandParams()
 		testParams.count = 1
 		testParams.schema.path = path
-		testParams.errOutput = io.Discard
+		testParams.errOutput = &buf
 
-		exitCode, err = opaTest([]string{regoFilePath}, testParams)
+		exitCode = opaTest([]string{regoFilePath}, testParams)
 	})
-	return exitCode, err
+
+	return exitCode, buf.Bytes()
 }
 func TestJSONSchemaSuccess(t *testing.T) {
 
@@ -1676,9 +1683,9 @@ test_p if {
 		"additionalProperties": false
    }`
 
-	_, err := testSchemasAnnotationWithJSONFile(regoContents, schema)
-	if err != nil {
-		t.Fatalf("unexpected error: %s", err)
+	errorCode, _ := testSchemasAnnotationWithJSONFile(regoContents, schema)
+	if errorCode != 0 {
+		t.Fatalf("unexpected error code: %d", errorCode)
 	}
 }
 
@@ -1715,11 +1722,13 @@ test_p if {
 "additionalProperties": false
 }`
 
-	exitCode, err := testSchemasAnnotationWithJSONFile(regoContents, schema)
+	exitCode, errOutput := testSchemasAnnotationWithJSONFile(regoContents, schema)
 	if exitCode == 0 {
 		t.Fatalf("didn't get expected error when schema is present and is defining a different type than being used.")
-	} else if !strings.Contains(err.Error(), "rego_type_error: match error") {
-		t.Fatalf("didn't get expected %s error when schema is defining a different type than being used; got: %v", ast.TypeErr, err)
+	}
+
+	if !strings.Contains(string(errOutput), "rego_type_error: match error") {
+		t.Fatalf("didn't get expected %s error when schema is defining a different type than being used; got: %v", ast.TypeErr, string(errOutput))
 	}
 }
 
@@ -1745,7 +1754,7 @@ test_p if {
 
 		done := make(chan struct{})
 		go func() {
-			_, _ = opaTest([]string{root}, testParams)
+			_ = opaTest([]string{root}, testParams)
 			<-done
 		}()
 
@@ -1763,7 +1772,10 @@ test_p if {
 		if err != nil {
 			t.Fatal(err)
 		}
-		f.Close()
+		err = f.Close()
+		if err != nil {
+			t.Fatal(err)
+		}
 
 		r := regexp.MustCompile(`FAIL \(.*s\)`)
 		expected = `%ROOT%/policy_test.rego:
@@ -1789,7 +1801,10 @@ Watching for changes ...
 			t.Fatal(err)
 		}
 
-		f.Close()
+		err = f.Close()
+		if err != nil {
+			t.Fatal(err)
+		}
 
 		expected = `PASS: 1/1
 ********************************************************************************
@@ -1852,7 +1867,7 @@ test_p {
 
 		done := make(chan struct{})
 		go func() {
-			_, _ = opaTest([]string{root}, testParams)
+			_ = opaTest([]string{root}, testParams)
 			<-done
 		}()
 
@@ -1870,7 +1885,10 @@ test_p {
 		if err != nil {
 			t.Fatal(err)
 		}
-		f.Close()
+		err = f.Close()
+		if err != nil {
+			t.Fatal(err)
+		}
 
 		r := regexp.MustCompile(`FAIL \(.*s\)`)
 		expected = `%ROOT%/policy_test.rego:
@@ -1896,7 +1914,10 @@ Watching for changes ...
 			t.Fatal(err)
 		}
 
-		f.Close()
+		err = f.Close()
+		if err != nil {
+			t.Fatal(err)
+		}
 
 		expected = `PASS: 1/1
 ********************************************************************************
@@ -1957,7 +1978,7 @@ test_p if {
 
 		done := make(chan struct{})
 		go func() {
-			_, _ = opaTest([]string{root}, testParams)
+			_ = opaTest([]string{root}, testParams)
 			<-done
 		}()
 
@@ -1975,7 +1996,10 @@ test_p if {
 		if err != nil {
 			t.Fatal(err)
 		}
-		f.Close()
+		err = f.Close()
+		if err != nil {
+			t.Fatal(err)
+		}
 
 		r := regexp.MustCompile(`FAIL \(.*s\)`)
 		expected = `%ROOT%/policy.rego:
@@ -2001,7 +2025,10 @@ Watching for changes ...
 			t.Fatal(err)
 		}
 
-		f.Close()
+		err = f.Close()
+		if err != nil {
+			t.Fatal(err)
+		}
 
 		expected = `PASS: 1/1
 ********************************************************************************
@@ -2041,7 +2068,7 @@ test_p if {
 
 		done := make(chan struct{})
 		go func() {
-			_, _ = opaTest([]string{root}, testParams)
+			_ = opaTest([]string{root}, testParams)
 			<-done
 		}()
 
@@ -2059,7 +2086,10 @@ test_p if {
 		if err != nil {
 			t.Fatal(err)
 		}
-		f.Close()
+		err = f.Close()
+		if err != nil {
+			t.Fatal(err)
+		}
 
 		r := regexp.MustCompile(`FAIL \(.*s\)`)
 		expected = `%ROOT%/policy.rego:
@@ -2084,7 +2114,10 @@ Watching for changes ...
 		if err != nil {
 			t.Fatal(err)
 		}
-		f.Close()
+		err = f.Close()
+		if err != nil {
+			t.Fatal(err)
+		}
 
 		expected = `PASS: 1/1
 ********************************************************************************
@@ -2165,7 +2198,7 @@ test_p if {
 
 				done := make(chan struct{})
 				go func() {
-					_, _ = opaTest([]string{root}, testParams)
+					_ = opaTest([]string{root}, testParams)
 					<-done
 				}()
 
@@ -2185,7 +2218,10 @@ test_p if {
 						t.Fatal(err)
 					}
 				}
-				f.Close()
+				err := f.Close()
+				if err != nil {
+					t.Fatal(err)
+				}
 
 				if !test.Eventually(t, 2*time.Second, func() bool {
 					expected := strings.ReplaceAll(tc.expectedOutput, "%ROOT%", root)
@@ -2197,11 +2233,14 @@ test_p if {
 
 				// write data to empty file
 				f, _ = os.OpenFile(path.Join(root, tc.fileName), os.O_RDWR|os.O_CREATE|os.O_TRUNC, 0644)
-				_, err := f.WriteString(tc.fixedFile)
+				_, err = f.WriteString(tc.fixedFile)
+				if err != nil {
+					t.Fatal(err)
+				}
+				err = f.Close()
 				if err != nil {
 					t.Fatal(err)
 				}
-				f.Close()
 
 				expected = "Watching for changes ..."
 				if !test.Eventually(t, 2*time.Second, func() bool {
@@ -2219,13 +2258,12 @@ test_p if {
 	}
 }
 
-func testExitCode(rego string, skipExitZero bool) (int, error) {
+func testExitCode(rego string, skipExitZero bool) int {
 	files := map[string]string{
 		"test.rego": rego,
 	}
 
 	var exitCode int
-	var err error
 	test.WithTempFS(files, func(path string) {
 		regoFilePath := filepath.Join(path, "test.rego")
 
@@ -2235,9 +2273,9 @@ func testExitCode(rego string, skipExitZero bool) (int, error) {
 		testParams.errOutput = io.Discard
 		testParams.output = io.Discard
 
-		exitCode, err = opaTest([]string{regoFilePath}, testParams)
+		exitCode = opaTest([]string{regoFilePath}, testParams)
 	})
-	return exitCode, err
+	return exitCode
 }
 
 func TestExitCode(t *testing.T) {
@@ -2314,7 +2352,7 @@ func TestExitCode(t *testing.T) {
 
 	for name, tc := range testCases {
 		t.Run(name, func(t *testing.T) {
-			exitCode, _ := testExitCode(tc.Test, tc.ExitZeroOnSkipped)
+			exitCode := testExitCode(tc.Test, tc.ExitZeroOnSkipped)
 
 			if exitCode != tc.ExpectedExitCode {
 				t.Errorf("Expected exit code to be %d but got %d", tc.ExpectedExitCode, exitCode)
@@ -2427,7 +2465,7 @@ Lines not covered:
 				testParams.count = 1
 				testParams.errOutput = &buf
 
-				exitCode, _ := opaTest([]string{root}, testParams)
+				exitCode := opaTest([]string{root}, testParams)
 				if exitCode != tc.expectedExitCode {
 					t.Fatalf("unexpected exit code: %d", exitCode)
 				}
@@ -2517,7 +2555,7 @@ test_l if {
 						}
 
 						testBundle := bundle.Bundle{
-							Data: map[string]interface{}{},
+							Data: map[string]any{},
 						}
 						for k, v := range tc.files {
 							testBundle.Modules = append(testBundle.Modules, bundle.ModuleFile{
@@ -2547,7 +2585,7 @@ test_l if {
 						paths = []string{root}
 					}
 
-					exitCode, _ := opaTest(paths, testParams)
+					exitCode := opaTest(paths, testParams)
 					if len(tc.expErrs) > 0 {
 						if exitCode == 0 {
 							t.Fatalf("expected non-zero exit code")
@@ -2668,6 +2706,27 @@ l2[v] {
 	v := l1[_]
 }
 
+test_l {
+	l1 == l2
+}`,
+			},
+			expErrs: []string{
+				"test.rego:4: rego_parse_error: `if` keyword is required before rule body",
+				"test.rego:4: rego_parse_error: `contains` keyword is required for partial set rules",
+				"test.rego:8: rego_parse_error: `if` keyword is required before rule body",
+			},
+		},
+		{
+			note:         "v0 module, not v0-compatible, v0 capabilities without rego_v1 feature",
+			capabilities: capsWithoutFeat(ast.RegoV0, ast.FeatureRegoV1),
+			files: map[string]string{
+				"/test.rego": `package test
+
+l1 := {1, 3, 5}
+l2[v] {
+	v := l1[_]
+}
+
 test_l {
 	l1 == l2
 }`,
@@ -2784,6 +2843,22 @@ l2 contains v if {
 	v := l1[_]
 }
 
+test_l if {
+	l1 == l2
+}`,
+			},
+		},
+		{
+			note:         "v1 module, not v0-compatible, v0 capabilities without rego_v1 feature",
+			capabilities: capsWithoutFeat(ast.RegoV0, ast.FeatureRegoV1),
+			files: map[string]string{
+				"/test.rego": `package test
+
+l1 := {1, 3, 5}
+l2 contains v if {
+	v := l1[_]
+}
+
 test_l if {
 	l1 == l2
 }`,
@@ -2827,7 +2902,7 @@ test_l if {
 						}
 
 						testBundle := bundle.Bundle{
-							Data: map[string]interface{}{},
+							Data: map[string]any{},
 						}
 						for k, v := range tc.files {
 							testBundle.Modules = append(testBundle.Modules, bundle.ModuleFile{
@@ -2859,7 +2934,7 @@ test_l if {
 						paths = []string{root}
 					}
 
-					exitCode, _ := opaTest(paths, testParams)
+					exitCode := opaTest(paths, testParams)
 					if len(tc.expErrs) > 0 {
 						if exitCode == 0 {
 							t.Fatalf("expected non-zero exit code")
@@ -3081,7 +3156,7 @@ test_l if {
 						}
 
 						testBundle := bundle.Bundle{
-							Data: map[string]interface{}{},
+							Data: map[string]any{},
 						}
 						for k, v := range tc.files {
 							testBundle.Modules = append(testBundle.Modules, bundle.ModuleFile{
@@ -3113,7 +3188,7 @@ test_l if {
 						paths = []string{root}
 					}
 
-					exitCode, _ := opaTest(paths, testParams)
+					exitCode := opaTest(paths, testParams)
 					if tc.expErr != "" {
 						if exitCode == 0 {
 							t.Fatalf("expected non-zero exit code")
@@ -3411,9 +3486,7 @@ test_l if {
 					if bundleType.tar {
 						files["bundle.tar.gz"] = ""
 					} else {
-						for k, v := range tc.files {
-							files[k] = v
-						}
+						maps.Copy(files, tc.files)
 					}
 
 					test.WithTempFS(files, func(root string) {
@@ -3445,7 +3518,7 @@ test_l if {
 						testParams.output = &buf
 						testParams.errOutput = &errBuf
 
-						exitCode, _ := opaTest([]string{p}, testParams)
+						exitCode := opaTest([]string{p}, testParams)
 						if tc.expErr != "" {
 							if exitCode == 0 {
 								t.Fatalf("expected non-zero exit code")
@@ -3489,12 +3562,127 @@ func TestTestBenchFailingTest(t *testing.T) {
 		tp.benchmark = true
 		tp.count = 1
 
-		exitCode, err := opaTest([]string{fp}, tp)
+		exitCode := opaTest([]string{fp}, tp)
 		if exitCode == 0 {
-			t.Fatalf("Expected exit code 0, got %d", exitCode)
-		}
-		if err != nil {
-			t.Fatalf("Unexpected error: %v", err)
+			t.Fatalf("Expected exit code != 0, got %d", exitCode)
 		}
 	})
 }
+
+func TestTestRunParallel(t *testing.T) {
+	tests := []struct {
+		note     string
+		parallel int
+	}{
+		{
+			note:     "default workers",
+			parallel: 0,
+		},
+		{
+			note:     "1 workers",
+			parallel: 1,
+		},
+		{
+			note:     "2 workers",
+			parallel: 2,
+		},
+		{
+			note:     "100 workers",
+			parallel: 100,
+		},
+	}
+
+	for _, tc := range tests {
+		testParams := newTestCommandParams()
+		testParams.parallel = tc.parallel
+
+		files := map[string]string{
+			"policy1.rego": `package test
+l1 := {1, 3, 5}
+l2 contains v if {
+	v := l1[_]
+}`,
+			"policy2.rego": `package test
+test_l if {
+	l1 == l2
+}`}
+
+		var exitCode int
+		test.WithTempFS(files, func(root string) {
+			exitCode = opaTest([]string{root}, testParams)
+		})
+
+		if exitCode > 0 {
+			t.Fatalf("unexpected exit code: %d", exitCode)
+		}
+	}
+}
+
+func TestWithDefaultRegoPlugin(t *testing.T) {
+	// We're injecting a default rego plugin that always returns true.
+	// If it's picked as a default (as intended), the tests run here will also
+	// yield true. If it wasn't picked, we'd use topdown, and would see a failing
+	// test.
+	tp := &testPlugin{}
+	rego.RegisterPlugin(targetPlugin, tp)
+	t.Cleanup(func() { tp.target = targetPlugin })
+
+	t.Run("test", func(t *testing.T) {
+		test.WithTempFS(map[string]string{"test.rego": "package test\ntest_true if false"}, func(path string) {
+			fp := filepath.Join(path, "test.rego")
+			tp := newTestCommandParams()
+			tp.output = io.Discard
+			tp.count = 1
+
+			exitCode := opaTest([]string{fp}, tp)
+			if exitCode != 0 {
+				t.Fatalf("Expected exit code 0, got %d", exitCode)
+			}
+		})
+	})
+	t.Run("eval", func(t *testing.T) {
+		params := newEvalCommandParams()
+		params.fail = true
+		query := "2+2 = 5" // unification will fail ("2+2 == 5" would be false, but defined)
+
+		defined, err := eval([]string{query}, params, io.Discard)
+		if err != nil {
+			t.Fatal("unexpected error", err)
+		}
+		if !defined {
+			t.Errorf("expected defined result")
+		}
+	})
+
+	t.Run("repl", func(t *testing.T) {
+		ctx := context.Background()
+		store := inmem.New()
+		var buffer bytes.Buffer
+		repl := repl.New(store, "", &buffer, "", 0, "")
+		if err := repl.OneShot(ctx, "2+2==5"); err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+		result := buffer.String()
+		if result != "true\n" {
+			t.Errorf("Expected result to be false but got: %v", result)
+		}
+	})
+}
+
+type testPlugin struct {
+	target string
+}
+
+func (t *testPlugin) IsTarget(tgt string) bool {
+	return tgt == t.target // t == "" makes it the global default
+}
+
+func (*testPlugin) PrepareForEval(context.Context, *ir.Policy, ...rego.PrepareOption) (rego.TargetPluginEval, error) {
+	return &testPlugin{}, nil
+}
+
+func (*testPlugin) Eval(context.Context, *rego.EvalContext, ast.Value) (ast.Value, error) {
+	return ast.NewSet(ast.NewTerm(ast.NewObject([2]*ast.Term{ast.StringTerm("^term1"), ast.BooleanTerm(true)}))), nil
+}
+
+const targetPlugin = "rego_test_default_plugin"
