@@ -11,6 +11,7 @@ import (
 	"errors"
 	"fmt"
 	"io"
+	"maps"
 	"os"
 	"path/filepath"
 	"reflect"
@@ -54,25 +55,30 @@ type Discovery struct {
 	manager              *plugins.Manager
 	config               *Config
 	factories            map[string]plugins.Factory
-	downloader           bundle.Loader                       // discovery bundle downloader
-	status               *bundle.Status                      // discovery status
-	listenersMtx         sync.Mutex                          // lock for listener map
-	listeners            map[interface{}]func(bundle.Status) // listeners for discovery update events
-	etag                 string                              // discovery bundle etag for caching purposes
+	downloader           bundle.Loader               // discovery bundle downloader
+	status               *bundle.Status              // discovery status
+	listenersMtx         sync.Mutex                  // lock for listener map
+	listeners            map[any]func(bundle.Status) // listeners for discovery update events
+	etag                 string                      // discovery bundle etag for caching purposes
 	metrics              metrics.Metrics
 	readyOnce            sync.Once
 	logger               logging.Logger
 	bundlePersistPath    string
 	hooks                hooks.Hooks
-	bootConfig           map[string]interface{}
+	bootConfig           map[string]any
 	overriddenConfigKeys []string
 }
 
 // Factories provides a set of factory functions to use for
-// instantiating custom plugins.
+// instantiating custom plugins. The passed map will be merged
+// with what's already on the `Discovery` instance, overwriting
+// existing keys on clashes.
 func Factories(fs map[string]plugins.Factory) func(*Discovery) {
 	return func(d *Discovery) {
-		d.factories = fs
+		if d.factories == nil {
+			d.factories = make(map[string]plugins.Factory, len(fs))
+		}
+		maps.Copy(d.factories, fs)
 	}
 }
 
@@ -89,7 +95,7 @@ func Hooks(hs hooks.Hooks) func(*Discovery) {
 	}
 }
 
-func BootConfig(bootConfig map[string]interface{}) func(*Discovery) {
+func BootConfig(bootConfig map[string]any) func(*Discovery) {
 	return func(d *Discovery) {
 		d.bootConfig = bootConfig
 	}
@@ -105,13 +111,15 @@ func New(manager *plugins.Manager, opts ...func(*Discovery)) (*Discovery, error)
 		f(result)
 	}
 
+	result.logger = manager.Logger().WithFields(map[string]any{"plugin": Name})
+
 	config, err := NewConfigBuilder().WithBytes(manager.Config.Discovery).WithServices(manager.Services()).
 		WithKeyConfigs(manager.PublicKeys()).Parse()
 
 	if err != nil {
 		return nil, err
 	} else if config == nil {
-		if _, err := getPluginSet(result.factories, manager, manager.Config, result.metrics, nil); err != nil {
+		if _, err := getPluginSet(result.factories, manager, manager.Config, result.metrics, result.logger, nil); err != nil {
 			return nil, err
 		}
 		return result, nil
@@ -141,7 +149,7 @@ func New(manager *plugins.Manager, opts ...func(*Discovery)) (*Discovery, error)
 		Name: Name,
 	}
 
-	result.logger = manager.Logger().WithFields(map[string]interface{}{"plugin": Name})
+	result.logger = manager.Logger().WithFields(map[string]any{"plugin": Name})
 
 	manager.UpdatePluginStatus(Name, &plugins.Status{State: plugins.StateNotReady})
 	return result, nil
@@ -177,7 +185,7 @@ func (c *Discovery) Stop(ctx context.Context) {
 }
 
 // Reconfigure is a no-op on discovery.
-func (*Discovery) Reconfigure(context.Context, interface{}) {
+func (*Discovery) Reconfigure(context.Context, any) {
 }
 
 // Lookup returns the discovery plugin registered with the manager.
@@ -202,19 +210,19 @@ func (c *Discovery) Trigger(ctx context.Context) error {
 	return c.downloader.Trigger(ctx)
 }
 
-func (c *Discovery) RegisterListener(name interface{}, f func(bundle.Status)) {
+func (c *Discovery) RegisterListener(name any, f func(bundle.Status)) {
 	c.listenersMtx.Lock()
 	defer c.listenersMtx.Unlock()
 
 	if c.listeners == nil {
-		c.listeners = map[interface{}]func(bundle.Status){}
+		c.listeners = map[any]func(bundle.Status){}
 	}
 
 	c.listeners[name] = f
 }
 
 // Unregister a listener to stop receiving status updates.
-func (c *Discovery) Unregister(name interface{}) {
+func (c *Discovery) Unregister(name any) {
 	c.listenersMtx.Lock()
 	defer c.listenersMtx.Unlock()
 
@@ -422,7 +430,7 @@ func (c *Discovery) applyLocalPluginConfigOverride(conf *config.Config) (*config
 		return nil, nil, err
 	}
 
-	var newConfig map[string]interface{}
+	var newConfig map[string]any
 	err = util.Unmarshal(raw, &newConfig)
 	if err != nil {
 		return nil, nil, err
@@ -509,7 +517,7 @@ func (c *Discovery) processBundle(ctx context.Context, b *bundleApi.Bundle) (*pl
 		return nil, err
 	}
 
-	ps, err := getPluginSet(c.factories, c.manager, overriddenConfig, c.metrics, c.config.Trigger)
+	ps, err := getPluginSet(c.factories, c.manager, overriddenConfig, c.metrics, c.logger, c.config.Trigger)
 	if err != nil {
 		return nil, err
 	}
@@ -565,7 +573,8 @@ func evaluateBundle(ctx context.Context, id string, info *ast.Term, b *bundleApi
 		return nil, err
 	}
 
-	return config.ParseConfig(bs, id)
+	processedConf := cfg.SubEnvVars(string(bs))
+	return config.ParseConfig([]byte(processedConf), id)
 }
 
 type pluginSet struct {
@@ -574,17 +583,17 @@ type pluginSet struct {
 }
 
 type pluginreconfig struct {
-	Config interface{}
+	Config any
 	Plugin plugins.Plugin
 }
 
 type pluginfactory struct {
 	name    string
 	factory plugins.Factory
-	config  interface{}
+	config  any
 }
 
-func getPluginSet(factories map[string]plugins.Factory, manager *plugins.Manager, config *config.Config, m metrics.Metrics, trigger *plugins.TriggerMode) (*pluginSet, error) {
+func getPluginSet(factories map[string]plugins.Factory, manager *plugins.Manager, config *config.Config, m metrics.Metrics, l logging.Logger, trigger *plugins.TriggerMode) (*pluginSet, error) {
 
 	// Parse and validate plugin configurations.
 	pluginNames := []string{}
@@ -628,7 +637,7 @@ func getPluginSet(factories map[string]plugins.Factory, manager *plugins.Manager
 	}
 
 	decisionLogsConfig, err := logs.NewConfigBuilder().WithBytes(config.DecisionLogs).WithServices(manager.Services()).
-		WithPlugins(pluginNames).WithTriggerMode(trigger).Parse()
+		WithPlugins(pluginNames).WithTriggerMode(trigger).WithLogger(l).Parse()
 	if err != nil {
 		return nil, err
 	}
@@ -743,7 +752,7 @@ func registerBundleStatusUpdates(m *plugins.Manager) {
 
 // mergeValuesAndListOverrides will merge source and destination map, preferring values from the source map.
 // It will also return a list of keys in the destination map which were overridden by those in the source map
-func mergeValuesAndListOverrides(dest map[string]interface{}, src map[string]interface{}, prefix string) (map[string]interface{}, []string) {
+func mergeValuesAndListOverrides(dest map[string]any, src map[string]any, prefix string) (map[string]any, []string) {
 	overriddenKeys := []string{}
 
 	for k, v := range src {
@@ -758,7 +767,7 @@ func mergeValuesAndListOverrides(dest map[string]interface{}, src map[string]int
 			fullKey = fmt.Sprintf("%v.%v", prefix, k)
 		}
 
-		nextMap, ok := v.(map[string]interface{})
+		nextMap, ok := v.(map[string]any)
 		// If it isn't another map, overwrite the value
 		if !ok {
 			if !reflect.DeepEqual(dest[k], v) {
@@ -768,7 +777,7 @@ func mergeValuesAndListOverrides(dest map[string]interface{}, src map[string]int
 			continue
 		}
 		// Edge case: If the key exists in the destination, but isn't a map
-		destMap, isMap := dest[k].(map[string]interface{})
+		destMap, isMap := dest[k].(map[string]any)
 		// If the source map has a map for this key, prefer it
 		if !isMap {
 			dest[k] = v
