@@ -14,6 +14,7 @@ import (
 	"fmt"
 	"io"
 	mr "math/rand"
+	"net/http"
 	"net/url"
 	"os"
 	"os/signal"
@@ -23,7 +24,7 @@ import (
 	"time"
 
 	"github.com/fsnotify/fsnotify"
-	"github.com/gorilla/mux"
+	prometheus_sdk "github.com/prometheus/client_golang/prometheus"
 	"go.opentelemetry.io/contrib/instrumentation/net/http/otelhttp"
 	"go.opentelemetry.io/otel/exporters/otlp/otlptrace"
 	"go.opentelemetry.io/otel/propagation"
@@ -43,6 +44,7 @@ import (
 	"github.com/open-policy-agent/opa/v1/ast"
 	"github.com/open-policy-agent/opa/v1/bundle"
 	opa_config "github.com/open-policy-agent/opa/v1/config"
+	"github.com/open-policy-agent/opa/v1/hooks"
 	"github.com/open-policy-agent/opa/v1/loader"
 	"github.com/open-policy-agent/opa/v1/logging"
 	"github.com/open-policy-agent/opa/v1/metrics"
@@ -205,6 +207,12 @@ type Params struct {
 	// SkipBundleVerification flag controls whether OPA will verify a signed bundle
 	SkipBundleVerification bool
 
+	// BundleActivatorPlugin controls the name of the activator plugin used to load bundles into the store.
+	BundleActivatorPlugin string
+
+	// BundleLazyLoadingMode flag controls whether OPA will load bundle contents in lazy mode.
+	BundleLazyLoadingMode bool
+
 	// SkipKnownSchemaCheck flag controls whether OPA will perform type checking on known input schemas
 	SkipKnownSchemaCheck bool
 
@@ -215,14 +223,17 @@ type Params struct {
 
 	// Router is the router to which handlers for the REST API are added.
 	// Router uses a first-matching-route-wins strategy, so no existing routes are overridden
-	// If it is nil, a new mux.Router will be created
-	Router *mux.Router
+	// If it is nil, a new http.ServeMux will be created
+	Router *http.ServeMux
 
 	// DiskStorage, if set, will make the runtime instantiate a disk-backed storage
 	// implementation (instead of the default, in-memory store).
 	// It can also be enabled via config, and this runtime field takes precedence.
 	DiskStorage *disk.Options
 
+	// StoreBuilder allows passing a storage backend builder
+	StoreBuilder func(_ context.Context, _ logging.Logger, _ prometheus_sdk.Registerer, config []byte, id string) (storage.Store, error)
+
 	DistributedTracingOpts tracing.Options
 
 	// Check if default Addr is set or the user has changed it.
@@ -249,6 +260,17 @@ type Params struct {
 	// Evaluation performance is affected in that data doesn't need to be converted to AST during evaluation.
 	// Only applicable when using the default in-memory store, and not when used together with the DiskStorage option.
 	ReadAstValuesFromStore bool
+
+	// ExtraDiscoveryOpts allows for passing options to the discovery plugin, as instantiated by the runtime.
+	ExtraDiscoveryOpts []func(*discovery.Discovery)
+
+	// Hooks is our generic extension mechanism.
+	Hooks hooks.Hooks
+
+	// NDBCacheEnabled allows enabling the non-deterministic builtin cache globally.
+	NDBCacheEnabled bool
+
+	Brand string
 }
 
 func (p *Params) regoVersion() ast.RegoVersion {
@@ -279,6 +301,7 @@ func NewParams() Params {
 		Output:             os.Stdout,
 		BundleMode:         false,
 		EnableVersionCheck: false,
+		Brand:              "OPA", // default
 	}
 }
 
@@ -300,7 +323,7 @@ type Runtime struct {
 	logger            logging.Logger
 	server            *server.Server
 	metrics           *prometheus.Provider
-	reporter          *report.Reporter
+	reporter          report.Reporter
 	traceExporter     *otlptrace.Exporter
 	loadedPathsResult *initload.LoadPathsResult
 
@@ -347,6 +370,10 @@ func NewRuntime(ctx context.Context, params Params) (*Runtime, error) {
 		logger = stdLogger
 	}
 
+	if err := params.Hooks.Validate(); err != nil {
+		return nil, err
+	}
+
 	var filePaths []string
 	urlPathCount := 0
 	for _, path := range params.Paths {
@@ -368,10 +395,10 @@ func NewRuntime(ctx context.Context, params Params) (*Runtime, error) {
 		return nil, fmt.Errorf("config error: %w", err)
 	}
 
-	var reporter *report.Reporter
+	var reporter report.Reporter
 	if params.EnableVersionCheck {
 		var err error
-		reporter, err = report.New(params.ID, report.Options{Logger: logger})
+		reporter, err = report.New(report.Options{Logger: logger})
 		if err != nil {
 			return nil, fmt.Errorf("config error: %w", err)
 		}
@@ -379,7 +406,7 @@ func NewRuntime(ctx context.Context, params Params) (*Runtime, error) {
 
 	regoVersion := params.regoVersion()
 
-	loaded, err := initload.LoadPathsForRegoVersion(regoVersion, params.Paths, params.Filter, params.BundleMode, params.BundleVerificationConfig, params.SkipBundleVerification, false, false, nil, nil)
+	loaded, err := initload.LoadPathsForRegoVersion(regoVersion, params.Paths, params.Filter, params.BundleMode, params.BundleVerificationConfig, params.SkipBundleVerification, params.BundleLazyLoadingMode, false, false, nil, nil)
 	if err != nil {
 		return nil, fmt.Errorf("load error: %w", err)
 	}
@@ -399,7 +426,7 @@ func NewRuntime(ctx context.Context, params Params) (*Runtime, error) {
 	}
 
 	if params.Router == nil {
-		params.Router = mux.NewRouter()
+		params.Router = http.NewServeMux()
 	}
 
 	metricsConfig, parseConfigErr := extractMetricsConfig(config, params)
@@ -416,12 +443,18 @@ func NewRuntime(ctx context.Context, params Params) (*Runtime, error) {
 		}
 	}
 
-	if params.DiskStorage != nil {
+	switch {
+	case params.DiskStorage != nil:
 		store, err = disk.New(ctx, logger, metrics, *params.DiskStorage)
 		if err != nil {
 			return nil, fmt.Errorf("initialize disk store: %w", err)
 		}
-	} else {
+	case params.StoreBuilder != nil:
+		store, err = params.StoreBuilder(ctx, logger, metrics, config, params.ID)
+		if err != nil {
+			return nil, fmt.Errorf("initialize store: %w", err)
+		}
+	default:
 		store = inmem.NewWithOpts(inmem.OptRoundTripOnWrite(false),
 			inmem.OptReturnASTValuesOnRead(params.ReadAstValuesFromStore))
 	}
@@ -455,6 +488,8 @@ func NewRuntime(ctx context.Context, params Params) (*Runtime, error) {
 		plugins.WithEnableTelemetry(params.EnableVersionCheck),
 		plugins.WithParserOptions(params.parserOptions()),
 		plugins.WithDistributedTracingOpts(params.DistributedTracingOpts),
+		plugins.WithBundleActivatorPlugin(params.BundleActivatorPlugin),
+		plugins.WithHooks(params.Hooks),
 	)
 	if err != nil {
 		return nil, fmt.Errorf("config error: %w", err)
@@ -470,13 +505,20 @@ func NewRuntime(ctx context.Context, params Params) (*Runtime, error) {
 		}
 	}
 
-	var bootConfig map[string]interface{}
+	var bootConfig map[string]any
 	err = util.Unmarshal(config, &bootConfig)
 	if err != nil {
 		return nil, fmt.Errorf("config error: %w", err)
 	}
 
-	disco, err := discovery.New(manager, discovery.Factories(registeredPlugins), discovery.Metrics(metrics), discovery.BootConfig(bootConfig))
+	opts := make([]func(*discovery.Discovery), 0, len(params.ExtraDiscoveryOpts)+3)
+	opts = append(opts,
+		discovery.Factories(registeredPlugins),
+		discovery.Metrics(metrics),
+		discovery.BootConfig(bootConfig),
+	)
+	opts = append(opts, params.ExtraDiscoveryOpts...)
+	disco, err := discovery.New(manager, opts...)
 	if err != nil {
 		return nil, fmt.Errorf("config error: %w", err)
 	}
@@ -532,7 +574,7 @@ func (rt *Runtime) ServerStatus() ServerStatus {
 }
 
 // StartServer starts the runtime in server mode. This function will block the
-// calling goroutine and will exit the program on error.
+// calling goroutine.
 func (rt *Runtime) StartServer(ctx context.Context) {
 	err := rt.Serve(ctx)
 	if err != nil {
@@ -557,7 +599,7 @@ func (rt *Runtime) Serve(ctx context.Context) error {
 		rt.Params.DiagnosticAddrs = &[]string{}
 	}
 
-	rt.logger.WithFields(map[string]interface{}{
+	rt.logger.WithFields(map[string]any{
 		"addrs":            *rt.Params.Addrs,
 		"diagnostic-addrs": *rt.Params.DiagnosticAddrs,
 	}).Info(serverInitializingMessage)
@@ -571,17 +613,17 @@ func (rt *Runtime) Serve(ctx context.Context) error {
 	// NOTE(tsandall): at some point, hopefully we can remove this because the
 	// Go runtime will just do the right thing. Until then, try to set
 	// GOMAXPROCS based on the CPU quota applied to the process.
-	undo, err := maxprocs.Set(maxprocs.Logger(func(f string, a ...interface{}) {
+	undo, err := maxprocs.Set(maxprocs.Logger(func(f string, a ...any) {
 		rt.logger.Debug(f, a...)
 	}))
 	if err != nil {
-		rt.logger.WithFields(map[string]interface{}{"err": err}).Debug("Failed to set GOMAXPROCS from CPU quota.")
+		rt.logger.WithFields(map[string]any{"err": err}).Debug("Failed to set GOMAXPROCS from CPU quota.")
 	}
 
 	defer undo()
 
 	if err := rt.Manager.Start(ctx); err != nil {
-		rt.logger.WithFields(map[string]interface{}{"err": err}).Error("Failed to start plugins.")
+		rt.logger.WithFields(map[string]any{"err": err}).Error("Failed to start plugins.")
 		return err
 	}
 
@@ -589,7 +631,7 @@ func (rt *Runtime) Serve(ctx context.Context) error {
 
 	if rt.traceExporter != nil {
 		if err := rt.traceExporter.Start(ctx); err != nil {
-			rt.logger.WithFields(map[string]interface{}{"err": err}).Error("Failed to start OpenTelemetry trace exporter.")
+			rt.logger.WithFields(map[string]any{"err": err}).Error("Failed to start OpenTelemetry trace exporter.")
 			return err
 		}
 	}
@@ -613,11 +655,13 @@ func (rt *Runtime) Serve(ctx context.Context) error {
 		WithMetrics(rt.metrics).
 		WithMinTLSVersion(rt.Params.MinTLSVersion).
 		WithCipherSuites(rt.Params.CipherSuites).
-		WithDistributedTracingOpts(rt.Params.DistributedTracingOpts)
+		WithDistributedTracingOpts(rt.Params.DistributedTracingOpts).
+		WithHooks(rt.Params.Hooks).
+		WithNDBCacheEnabled(rt.Params.NDBCacheEnabled)
 
 	// If decision_logging plugin enabled, check to see if we opted in to the ND builtins cache.
 	if lp := logs.Lookup(rt.Manager); lp != nil {
-		rt.server = rt.server.WithNDBCacheEnabled(rt.Manager.Config.NDBuiltinCacheEnabled())
+		rt.server = rt.server.WithNDBCacheEnabled(rt.Params.NDBCacheEnabled || rt.Manager.Config.NDBuiltinCacheEnabled())
 	}
 
 	if rt.Params.DiagnosticAddrs != nil {
@@ -648,13 +692,13 @@ func (rt *Runtime) Serve(ctx context.Context) error {
 	defer cancel()
 	rt.server, err = rt.server.Init(ctx)
 	if err != nil {
-		rt.logger.WithFields(map[string]interface{}{"err": err}).Error("Unable to initialize server.")
+		rt.logger.WithFields(map[string]any{"err": err}).Error("Unable to initialize server.")
 		return err
 	}
 
 	if rt.Params.Watch {
 		if err := rt.startWatcher(ctx, rt.Params.Paths, rt.onReloadLogger); err != nil {
-			rt.logger.WithFields(map[string]interface{}{"err": err}).Error("Unable to open watch.")
+			rt.logger.WithFields(map[string]any{"err": err}).Error("Unable to open watch.")
 			return err
 		}
 	}
@@ -678,13 +722,13 @@ func (rt *Runtime) Serve(ctx context.Context) error {
 	if err := rt.waitPluginsReady(
 		100*time.Millisecond,
 		time.Second*time.Duration(rt.Params.ReadyTimeout)); err != nil {
-		rt.logger.WithFields(map[string]interface{}{"err": err}).Error("Failed to wait for plugins activation.")
+		rt.logger.WithFields(map[string]any{"err": err}).Error("Failed to wait for plugins activation.")
 		return err
 	}
 
 	loops, err := rt.server.Listeners()
 	if err != nil {
-		rt.logger.WithFields(map[string]interface{}{"err": err}).Error("Unable to create listeners.")
+		rt.logger.WithFields(map[string]any{"err": err}).Error("Unable to create listeners.")
 		return err
 	}
 
@@ -719,7 +763,7 @@ func (rt *Runtime) Serve(ctx context.Context) error {
 		case <-signalc:
 			return rt.gracefulServerShutdown(rt.server)
 		case err := <-errc:
-			rt.logger.WithFields(map[string]interface{}{"err": err}).Error("Listener failed.")
+			rt.logger.WithFields(map[string]any{"err": err}).Error("Listener failed.")
 			os.Exit(1) //nolint:gocritic
 		}
 	}
@@ -750,10 +794,10 @@ func (rt *Runtime) DiagnosticAddrs() []string {
 }
 
 // StartREPL starts the runtime in REPL mode. This function will block the calling goroutine.
-func (rt *Runtime) StartREPL(ctx context.Context) {
+func (rt *Runtime) StartREPL(ctx context.Context) error {
 	if err := rt.Manager.Start(ctx); err != nil {
 		fmt.Fprintln(rt.Params.Output, "error starting plugins:", err)
-		os.Exit(1)
+		return err
 	}
 
 	defer rt.Manager.Stop(ctx)
@@ -767,7 +811,7 @@ func (rt *Runtime) StartREPL(ctx context.Context) {
 	if rt.Params.Watch {
 		if err := rt.startWatcher(ctx, rt.Params.Paths, onReloadPrinter(rt.Params.Output)); err != nil {
 			fmt.Fprintln(rt.Params.Output, "error opening watch:", err)
-			os.Exit(1) //nolint:gocritic
+			return err
 		}
 	}
 
@@ -778,7 +822,7 @@ func (rt *Runtime) StartREPL(ctx context.Context) {
 	}
 
 	rt.repl = repl
-	repl.Loop(ctx)
+	return repl.Loop(ctx)
 }
 
 // SetDistributedTracingLogging configures the distributed tracing's ErrorHandler,
@@ -804,19 +848,19 @@ func (rt *Runtime) checkOPAUpdateLoopDurations(ctx context.Context, done chan st
 	for {
 		resp, err := rt.reporter.SendReport(ctx)
 		if err != nil {
-			rt.logger.WithFields(map[string]interface{}{"err": err}).Debug("Unable to send OPA version report.")
+			rt.logger.WithFields(map[string]any{"err": err}).Debug("Unable to send %s version report.", rt.Params.Brand)
 		} else {
 			if resp.Latest.OPAUpToDate {
-				rt.logger.WithFields(map[string]interface{}{
+				rt.logger.WithFields(map[string]any{
 					"current_version": version.Version,
-				}).Debug("OPA is up to date.")
+				}).Debug("%s is up to date.", rt.Params.Brand)
 			} else {
-				rt.logger.WithFields(map[string]interface{}{
+				rt.logger.WithFields(map[string]any{
 					"download_opa":    resp.Latest.Download,
 					"release_notes":   resp.Latest.ReleaseNotes,
 					"current_version": version.Version,
 					"latest_version":  strings.TrimPrefix(resp.Latest.LatestRelease, "v"),
-				}).Info("OPA is out of date.")
+				}).Info("%s is out of date.", rt.Params.Brand)
 			}
 		}
 		select {
@@ -871,7 +915,7 @@ func (rt *Runtime) readWatcher(ctx context.Context, watcher *fsnotify.Watcher, p
 		removalMask := fsnotify.Remove | fsnotify.Rename
 		mask := fsnotify.Create | fsnotify.Write | removalMask
 		if (evt.Op & mask) != 0 {
-			rt.logger.WithFields(map[string]interface{}{
+			rt.logger.WithFields(map[string]any{
 				"event": evt.String(),
 			}).Debug("Registered file event.")
 			t0 := time.Now()
@@ -886,8 +930,7 @@ func (rt *Runtime) readWatcher(ctx context.Context, watcher *fsnotify.Watcher, p
 }
 
 func (rt *Runtime) processWatcherUpdate(ctx context.Context, paths []string, removed string) error {
-
-	return pathwatcher.ProcessWatcherUpdateForRegoVersion(ctx, rt.Manager.ParserOptions().RegoVersion, paths, removed, rt.Store, rt.Params.Filter, rt.Params.BundleMode, func(ctx context.Context, txn storage.Transaction, loaded *initload.LoadPathsResult) error {
+	return pathwatcher.ProcessWatcherUpdateForRegoVersion(ctx, rt.Manager.ParserOptions().RegoVersion, paths, removed, rt.Store, rt.Params.Filter, rt.Params.BundleMode, rt.Params.BundleLazyLoadingMode, func(ctx context.Context, txn storage.Transaction, loaded *initload.LoadPathsResult) error {
 		_, err := initload.InsertAndCompile(ctx, initload.InsertAndCompileOptions{
 			Store:         rt.Store,
 			Txn:           txn,
@@ -903,7 +946,7 @@ func (rt *Runtime) processWatcherUpdate(ctx context.Context, paths []string, rem
 
 func (rt *Runtime) getBanner() string {
 	var buf bytes.Buffer
-	fmt.Fprintf(&buf, "OPA %v (commit %v, built at %v)\n", version.Version, version.Vcs, version.Timestamp)
+	fmt.Fprintf(&buf, "%s %v (commit %v, built at %v)\n", rt.Params.Brand, version.Version, version.Vcs, version.Timestamp)
 	fmt.Fprintf(&buf, "\n")
 	fmt.Fprintf(&buf, "Run 'help' to see a list of commands and check for updates.\n")
 	return buf.String()
@@ -920,7 +963,7 @@ func (rt *Runtime) gracefulServerShutdown(s *server.Server) error {
 	defer cancel()
 	err := s.Shutdown(ctx)
 	if err != nil {
-		rt.logger.WithFields(map[string]interface{}{"err": err}).Error("Failed to shutdown server gracefully.")
+		rt.logger.WithFields(map[string]any{"err": err}).Error("Failed to shutdown server gracefully.")
 		return err
 	}
 	rt.logger.Info("Server shutdown.")
@@ -928,7 +971,7 @@ func (rt *Runtime) gracefulServerShutdown(s *server.Server) error {
 	if rt.traceExporter != nil {
 		err = rt.traceExporter.Shutdown(ctx)
 		if err != nil {
-			rt.logger.WithFields(map[string]interface{}{"err": err}).Error("Failed to shutdown OpenTelemetry trace exporter gracefully.")
+			rt.logger.WithFields(map[string]any{"err": err}).Error("Failed to shutdown OpenTelemetry trace exporter gracefully.")
 		}
 	}
 	return nil
@@ -955,7 +998,7 @@ func (rt *Runtime) waitPluginsReady(checkInterval, timeout time.Duration) error
 }
 
 func (rt *Runtime) onReloadLogger(d time.Duration, err error) {
-	rt.logger.WithFields(map[string]interface{}{
+	rt.logger.WithFields(map[string]any{
 		"duration": d,
 		"err":      err,
 	}).Info("Processed file watch event.")
@@ -968,7 +1011,7 @@ func (rt *Runtime) getWatcher(rootPaths []string) (*fsnotify.Watcher, error) {
 	}
 
 	for _, path := range watcher.WatchList() {
-		rt.logger.WithFields(map[string]interface{}{"path": path}).Debug("watching path")
+		rt.logger.WithFields(map[string]any{"path": path}).Debug("watching path")
 	}
 
 	return watcher, nil
@@ -993,8 +1036,8 @@ func urlPathToConfigOverride(pathCount int, path string) ([]string, error) {
 	}, nil
 }
 
-func errorLogger(logger logging.Logger) func(attrs map[string]interface{}, f string, a ...interface{}) {
-	return func(attrs map[string]interface{}, f string, a ...interface{}) {
+func errorLogger(logger logging.Logger) func(attrs map[string]any, f string, a ...any) {
+	return func(attrs map[string]any, f string, a ...any) {
 		logger.WithFields(attrs).Error(f, a...)
 	}
 }
